# test-jit.R: tests for the R jitter
#
# RA_TODO integrate this into check-develop

# dummy function for testing with standard R
#
# jit <- function(jit=NA, show=0)
# {
#     if (length(grep("ra", R.version.string)) > 0)
#         stop("Cannot redefine \"jit()\" when running ra")
#     c(jit, show, jit, jit)
# }

gctorture(0)
xglobal <- c(10,20,NaN,NA,0)
jit.flag <- 0
trace.flag <- 0

if (length(grep("ra", R.version.string)) != 1)
   stop("No \"ra\" in R.version.string")

show.diffs <- function(a,b)
{
    if(is.matrix(a)) {
        for(i in 1:nrow(a))
            for(j in 1:ncol(a))
                if(!identical(a[i,j], b[i,j]))
                    cat("        ans[", i, ",", j, "] <- ", a[i,j],
                        " # was ", b[i,j], "\n", sep="")
    } else if(is.vector(a)) {
        for(i in 1:length(a))
            if(!identical(a[i], b[i]))
                cat("        ans[", i, "] <- ", a[i],
                    " # was ", b[i], "\n ", sep="")
    } else {
        cat("non jitted answer\n")
        print.default(a)
        cat("jitted answer\n")
        print.default(b)
    }
}

test <- function(test.func, ...)
{
    jit.flag <<- 0; a <- test.func(...)

    # we set the trace.flag so we can see compilation results (to test
    # that compilation is indeed taking place)

    jit.flag <<- 1; trace.flag <<- 1; b <- test.func(...)

    if(!identical(b, a)) {
        show.diffs(a,b)
        stop("test ", paste(substitute(test.func), substitute(...)),
             " failed for jit(1,1)")
    }
}

#------------------------------------------------------------------------------
# basic tests

test.test <- function(...)  # test test() and jit()
{
    jit(jit.flag, trace.flag)
    if(jit.flag) {
        stopifnot(jit()[1] == jit.flag)
        stopifnot(jit()[2] == trace.flag)
        # can't use stopifnot directly here because then jit() will
        # be evaluated in a function and therefore will be false
        if(jit()[4] != jit.flag)
            stop("jit()[4] != jit.flag")
        stopifnot(jit()[4] == 0)
    } else {
        stopifnot(jit()[1] == 0)
        stopifnot(jit()[2] == 0)
        stopifnot(jit()[3] == 0)
        if(jit()[4] != 0)
            stop("jit()[4] != 0")
        stopifnot(jit()[4] == 0)
    }
    if(jit()[1])
        cat("jit is on\n")
    else
        cat("jit is off\n")
    1
}
test(test.test)
f010 <- function(...)   # simple assignment
{
    jit(jit.flag, trace.flag)
    print(jit())
    x <- 0
    for(i in as.double(1:3))
        x <- 3
    x
}
test(f010)
f012 <- function(...)   # assignment real1 <- real1
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in as.double(1:3))
        x <- i
    x
}
test(f012)
f013_global=7
f013 <- function(...)   # assignment real1 <- globalreal1
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
        x <- f013_global
    x
}
test(f013)
f014_global<-7
f014 <- function(...)   # assignment globalreal1 <- real1
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
        f014_global <- x
    f014_global
}
test(f014)
f015 <- function(...)   # assignment real1 <- globalreal1
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
        x <- f013_global + x + 9
    x
}
test(f015)
# f016_global<-7
# f016 <- function(...)   # assignment globalreal1 <- real1
# {
#     jit(jit.flag, trace.flag)
#     x <- 0
#     for(i in 1:3)
#         f016_global <- 3 + x * f016_global
#     f016_global
# }
# test(f016) # correctly causes "Cannot change jitted symbol 'f016_global' to local scope"
f018_global<-7
f018 <- function(...)   # assignment globalreal1 <- real1
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
        f018_global <- 3 + x
    f018_global
}
test(f018)
f020 <- function(...)   # assignment pi <- real1
{
    jit(jit.flag, trace.flag)
    saved.pi <- pi
    x <- 2
    for(i in 1:3)
        pi <- 1000 + x
    new.pi <- pi
    pi <- saved.pi
    new.pi
}
test(f020)
f021 <- function(...)   # assignment int1 <- int1
{
    jit(jit.flag, trace.flag)
    x <- as.integer(0)
    for(i in 1:3)
        x <- i
    x
}
test(f021)
f022 <- function(...)   # assignment real <- real
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    y <- c(-9,0,10,11,-12)
    for(i in 1:3)
        x <- y
    x
}
test(f022)
# test assignments of different types to vars of different scopes
f023_1 <- 4; f023_2 <- "a"; f023_3 <- NA; f023_4 <- 2:4
f023_5 <- 2; f023_6 <- 3;   f023_7 <- 8;  f023_8 <- -1
f023_9 <- -2
f023 <- function()
{
    nested_f <- function(x) x

    jit(jit.flag, trace.flag)

    x11 <- 4; x12 <- "a"; x13 <- NA; x14 <- 3:5
    x15 <- 2; x16 <- 3;   x17 <- 8;  x18 <- nested_f(as.integer(-1));
    x19 <- sin(-2)
    for(i in 1:3) {
        x1 <- 3;              x2 <- "b";               x3 <- TRUE;     x4 <- 4:6
        x5 <- atan2(2,3);     x6 <- besselI(2, nu=3);  x7 <- 7;        x8 <- nested_f(i)
        x9 <- sin(i)
        x11 <- 3;             x12 <- "b";              x13 <- TRUE;    x14 <- 5:7
        x15 <- atan2(2,3);    x16 <- besselI(2, nu=3); x17 <- 7;       x18 <- nested_f(i)
        x19 <- sin(i)
        f023_1 <- 3;          f023_2 <- "b";              f023_3 <- TRUE; f023_4 <- 6:8
        f023_5 <- atan2(2,3); f023_6 <- besselI(2, nu=3); f023_7 <- 7;    f023_8 <- nested_f(i)
        f023_9 <- sin(i)
    }
    c(x1,x2,x3,x4,
      x5,x6,x7,x8,
      x9,
      x11,x12,x13,x14,
      x15,x16,x17,x18,
      x19,
      f023_1,f023_2,f023_3,f023_4,
      f023_5, f023_6,f023_7,f023_8,
      f023_9)
}
f023()
# as f023 above but use nested funcs and args
f024_x <- -2;  f024_0a <- -1; f024_0b <- -9; f024_0c <- -8; f024_1 <- 4
f024_2 <- "a"; f024_3 <- NA;  f024_4 <- 2:4; f024_5 <- 2;   f024_6 <- 3
f024_7 <- 8;   f024_8 <- -1;  f024_9 <- -2
f024_aux <- function(y1, y2)
{
    f024_x <- -3; x20a <- y1; x20a <- y2; x20c <- y2; x21 <- 4
    x22 <- "a";   x23 <- NA;  x24 <- 3:5; x25 <- 2;   x26 <- 3
    x27 <- 8;     x28 <- -1;  x29 <- sin(-2)

    nested_f <- function(x) { x+1 }

    nested_g <- function(x) {

        if(trace.flag == 1)
            trace.flag <- 2      # show compilation details
        jit(jit.flag, trace.flag)

        x10a <- x;  xx10b <- y1;        x10c <- y2; x11 <- 4;
        x12 <- "b"; x13 <- NA;          x14 <- 3:5; x15 <- 2; x16 <- 3;
        x17 <- 8;  x18 <- nested_f(-1); x19 <- sin(-2)
        for(i in 1:3) {
            f024_x <- -4;      x0a <- x;
            x0b <- y1;         x0c <- y2; x1 <- 3
            x2 <- "c";         x3 <- TRUE;
            x4 <- 4:6;         x5 <- atan2(2,3);  x6 <- besselI(2, nu=3)
            x7 <- 7;           x8 <- nested_f(i + 1);
            x9 <- sin(i)
            x10a <- x + 1;     x10b <- y1 + 1;
            x10c <- y2 + 1;    x11 <- 3;  x12 <- "b"
            x13 <- TRUE;       x14 <- 5:7;
            x15 <- atan2(2,3); x16 <- besselI(2, nu=3)
            x17 <- 7;          x18 <- nested_f(as.double(i));
            x19 <- sin(i)
            x20a <- x + 2;     x20b <- y1 + 2;
            x20c <- y2 + 2;    x21 <- 4
            x22 <- "d";        x23 <- NA;
            x24 <- 7:9;        x25 <- atan2(1,2); x26 <- besselI(2, nu=3)
            x27 <- 8;          x28 <- nested_f(i-3);
            x29 <- sin(i+1)
            f024_0a <- x + 2;  f024_0b <- y1 + 2;
            f024_0c <- y2 + 2; f024_1 <- 3;    f024_2 <- "b";
            f024_3 <- TRUE;    f024_4 <- 6:8;
            f024_5 <- atan2(2,3); f024_6 <- besselI(2, nu=3)
            f024_7 <- 7;       f024_8 <- nested_f(as.double(i)); f024_9 <- sin(i)
        }
        c(x0a,x0b,x0c,x1,
          x2,x3,x4,x5,x6,
          x7,x8,x9,
          x10a,x10b,x10c,x11,
          x12,x13,x14,x15,x16,
          x17,x18,x19,
          x20a,x20b,x20c,x21,x22,x23,x24,x25,x26,
          x27,x28,x29,
          f024_x,f024_0a,f024_0b,f024_0c,f024_1,
          f024_2,f024_3,f024_4,f024_5,f024_6,
          f024_7,f024_8,f024_9)
   }
   nested_g(3.3)
}
f024 <- function(xparam)
{
    f024_aux(xparam, 3.2)
}
test(f024, 3.1)
f025 <- function(...)   # multiple assignments
{
    jit(jit.flag, trace.flag)
    x1 <- 0; x2 <- 1; x3 <- 2; x4 <- 3; x5 <- 4; x6 <- 5; x7 <- 6
    for(i in 1:3) {
        x1 <- x2 <- 90
        x3 <- x2
        x4 <- x1
        x1 <- x2 <- x5
        x5 <- 91
        x6 <- x7 <- x1 * x2
    }
    c(x1,x2,x3,x4,x5,x6,x7)
}
test(f025)

# RA_TODO must add tests for other combinations of assignments

f026 <- function(...) # basic binop1 with scalars
{
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 11:13)
        x <- i+1
    x
}
test(f026)
f028 <- function(...) # basic binop with vector and scalar
{                     # also test '=' versus '<-'
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 21:24)
        x = xglobal+1
    x
}
test(f028)
f026 <- function(...) # complicated expression
{
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 1:3)
        x <- 2 * ((x + 3) / (4 - x) + 5) + sin(12 * ((x * 13) - (1.1 ^ x) / 15))
    x
}
f_jiton_off <- function()   # test jit on, jit off
{
           x1 <- 3; for(i in 1:3) x1 <- x1 * 2
jit(1,1); x2 <- 3; for(i in 1:3) x2 <- x2 * 2
jit(1,2); x3 <- 3; for(i in 1:3) x3 <- x3 * 2
jit(1,3); x4 <- 3; for(i in 1:3) x4 <- x4 * 2
jit(1,1); x5 <- 3; for(i in 1:3) x5 <- x5 * 2
jit(0);   x6 <- 3; for(i in 1:3) x6 <- x6 * 2
jit(1,1); x7 <- 3; for(i in 1:3) x7 <- x7 * 2
           x8 <- 3; for(i in 1:3) x8 <- x8 * 2  # jit should still be on

if(all(x2) != all(x1))
    stop("test failed")
if(all(x3) != all(x1))
    stop("test failed")
if(all(x4) != all(x1))
    stop("test failed")
if(all(x5) != all(x1))
    stop("test failed")
if(all(x6) != all(x1))
    stop("test failed");
if(all(x7) != all(x1))
    stop("test failed");
if(all(x8) != all(x1))
    stop("test failed");
}
f_jiton_off()

# make sure jit is off even though we left it on in f_jiton_off
if(jit()[1]) stop("test failed: jit()[1] after f_jiton_jitoff")

f030 <- function(...) # test multiple exps in for block
{
    jit(jit.flag, trace.flag)
    x <- 5
    for(i in 1:3) {
        x <- x + 1
        x <- x + 2
    }
    x
}
test(f030)
f031 <- function(...) # simple subscript of local
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3) {
        x[i] <- x[i] + 99
    }
    x
}
test(f031)
f032 <- function(...) # simple subscript of global
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3) {
        x[i] <- xglobal[i] + 99
    }
    x
}
test(f032)
f034 <- function(...) # test multiple exps in for block
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3) {
        x[i] <- xglobal[i] + 99
        x[i+1] <- x[i] * 2
    }
    x
}
test(f034)
f040 <- function(...) # while
{
    jit(jit.flag, trace.flag)
    x <- 0
    while(x < 3)
        x <- x + 1
    x
}
test(f040)
f042 <- function(...)  # repeat and break
{
    jit(jit.flag, trace.flag)
    x <- 0
    repeat {
        x <- x + 1
        if(x > 30)
            break;
        x <- x + 10
    }
    x
}
test(f042)
f043 <- function(...)  # break and next
{
    jit(jit.flag, trace.flag)
    x <- 0
    repeat {
        x <- x + 1
        if(x == 3)
            next;
        x <- x + 1
        if(x == 7)
            break;
    }
    x
}
test(f043)
f043a <- function(...) # nested for loops
{                      # this tests check_jit_binding_value val != R_NilValue
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
        for(j in 1:3)
            x <- i+j
    x
}
test(f043a)
f044 <- function(...)  # nested for with ifs
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3)
       for(i in 1:3)
          x <- x + 1
    y <- 0
    for(i in 1:3) {
       for(i in 1:3) {
          y <- y + 1
          if(y == 15)
              break;
       }
    }
    z <- 0
    for(i in 1:3) {
       if(z > 15)
          break;
       for(i in 1:3) {
          z <- z + 1
       }
    }
    c(x,y,z)
}
test(f044)
f045 <- function(...)   # longish loop
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:10000)
        x <- x + .1
    x
}
test(f045)
f050 <- function(...) # large jitted expression, compare to f612
{
    if (trace.flag == 1)
        trace.flag = 2; 	# use trace=2 so we see "did not compile" msgs
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3) {
        x <- x +  1; x <- x +  2; x <- x +  3; x <- x +  4; x <- x +  5;
        x <- x +  6; x <- x +  7; x <- x +  8; x <- x +  9; x <- x + 10;
        x <- x + 11; x <- x + 12; x <- x + 13; x <- x + 14; x <- x + 15;
        x <- x + 16; x <- x + 17; x <- x + 18; x <- x + 19; x <- x + 20;
        x <- x + 21; x <- x + 22; x <- x + 23; x <- x + 24; x <- x + 25;
        x <- x + 26; x <- x + 27; x <- x + 28; x <- x + 29; x <- x + 30;
        x <- x + 31; x <- x + 32; x <- x + 33; x <- x + 34; x <- x + 35;
        x <- x + 36; x <- x + 37; x <- x + 38; x <- x + 39; x <- x + 40;
        x <- x + 41; x <- x + 42; x <- x + 43; x <- x + 44; x <- x + 45;
        x <- x + 46; x <- x + 47; x <- x + 48; x <- x + 49; x <- x + 50;
        x <- x + 51; x <- x + 52; x <- x + 53; x <- x + 54; x <- x + 55;
        x <- x + 56; x <- x + 57; x <- x + 58; x <- x + 59; x <- x + 60;
        x <- x + 61; x <- x + 62; x <- x + 63; x <- x + 64; x <- x + 65;
        x <- x + 66; x <- x + 67; x <- x + 68; x <- x + 69; x <- x + 70;
        x <- x + 71; x <- x + 72; x <- x + 73; x <- x + 74; x <- x + 75;
        x <- x + 76; x <- x + 77; x <- x + 78; x <- x + 79; x <- x + 80;
        x <- x + 81; x <- x + 82; x <- x + 83; x <- x + 84; x <- x + 85;
        x <- x + 86; x <- x + 87; x <- x + 88; x <- x + 89; x <- x + 90;
        x <- x + 91; x <- x + 92; x <- x + 93; x <- x + 94; x <- x + 95;
        x <- x + 96; x <- x + 97; x <- x + 98; x <- x + 99; x <- x + 100;
    }
    x
}
test(f050)
f060 <- function(...) # parentheses
{
    jit(jit.flag, trace.flag)
    x <- 0
    y <- c(-9,0,10,11,-12)
    for(i in 1:3) {
        x <- (i + 1)
        x <- ((i + 1)) + x
        x <- x + (i + 1) * (i + 2)
        x <- x + ((i * 2) + (i * 3))
        y[i] <- x + 3
        x <- x + y[i] + (y[(i+0) + 1]) * ((y[(i+1)]) + 2)
    }
}
test(f060)
f062 <- function(...) # multiple assignments
{
    jit(jit.flag, trace.flag)
    x <- 0
    y <- 1
    for(i in 1:3) {
        x <- y <- i + 1
        x <- x + 1
        y <- x + 1
    }
}
test(f062)
f064 <- function() { # test using jit(0) inside a loop in a jit block
    jit(jit.flag, 2)
    x <- 11
    for (i in 1:5) {
       x = i + x
       if (i > 1) {
           jit(0, 2)
           x = c(x, 3) # check that we can now change length of x
       }
       x = x + i
    }
    x
}
test(f064)
f070 <- function(...) # if(needed for next test)
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:5) {
        if(i %% 3 == 1)
            x[i] <- xglobal[i] + 10
        else if(i == 2)
            x[i] <- xglobal[i] + 20
        else if(i == 3)
            x[i] <- xglobal[i] + 30
    }
    x
}
test(f070)
f072 <- function(...) # basic NAMED handling
{
    jit(jit.flag, trace.flag)
    x <- 1
    y <- 2
    for(i in 1:5) {     # enable jitting
        x <- 3
        y <- x          # after this, x is NAMED
        x <- 4          # this should cause x to be copied before assignment
    }
    c(x, y)
}
test(f072)

# RA_TODO add more NAMED handling tests

#------------------------------------------------------------------------------
# jit state nesting

fa <- function()
{
    jit(1,1)
    x <- 1
    for(i in 1:3)
        x <- x + 1
    x
}
stopifnot(jit()[1] == 0)
fa()
stopifnot(jit()[1] == 0)
f_with_return <- function()
{
    jit(1,1)
    x <- 1
    for(i in 1:3)
        x <- x + 2
    return(x)
    for(i in 1:3)
        x <- x + 3
    x
}
f_with_return()
stopifnot(jit()[1] == 0)
f_call_fa <- function()
{
    x <- 1
    for(i in 1:3)
        x <- x + 4
    fa()
    stopifnot(jit()[1] == 0)
    stopifnot(jit()[2] == 0)
    stopifnot(jit()[3] == 0)
    for(i in 1:3)
        x <- x + 5
    x
}
f_call_fa()
stopifnot(jit()[1] == 0)
fb_call_fa <- function()
{
    jit(1,1)
    x <- 1
    for(i in 1:3)
        x <- x + 6
    cat("Expect a warning below: ignored jit(1,1) in fa because ...\n")
    fa()
    stopifnot(jit()[1] == 1)
    stopifnot(jit()[2] == 1)
    stopifnot(jit()[3] == 1)
    for(i in 1:3)
        x <- x + 7
    return(x);
    for(i in 1:3)
        x <- x + 8
    x
}
fb_call_fa()
stopifnot(jit()[1] == 0)
fb_call_fa_with_return <- function()
{
    jit(1,1)
    x <- 1
    for(i in 1:3)
        x <- x + 9
    cat("Expect a warning below: ignored jit(1,1) in f_with-return because ...\n")
    f_with_return()
    stopifnot(jit()[1] == 1)
    stopifnot(jit()[2] == 1)
    stopifnot(jit()[3] == 1)
    for(i in 1:3)
        x <- x + 10
    return(x);
    for(i in 1:3)
        x <- x + 11
    x
}
fb_call_fa_with_return()
stopifnot(jit()[1] == 0)
f098 <- function(...) # intermingle function calls
{
    foo <- function(x) x+99
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:5) {
        x <- x + sin(i) + 2 + 3
        x <- x + foo(i)
        x <- x + 4 + foo(x)
        # RA_TODO: jitter limitation in expression below, 6+7 not compiled
        x <- x + 5 + foo(x) + 6 + 7 
    }
}
test(f098)

#------------------------------------------------------------------------------
# arguments, promises

f200 <- function(arg)   # test arg passing
{
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 1:3)
        x <- x + arg
    x
}
test(f200, 1)
test(f200, 2)
f210 <- function(arg=9) # test default arg
{
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 1:3)
        x <- x + arg
    x
}
test(f210)
f212 <- function(arg=9) # test default arg
{
    jit(jit.flag, trace.flag)
    for(i in 1:3)
        arg <- arg + i
    arg
}
test(f212, 4)
test(f212)
aglobal <- 13
bglobal <- 14
f220 <- function(x)    # lexical scoping 1
{
    jit(jit.flag, trace.flag)
    a <- 9
    b <- 9.1
    bglobal <- 9.2
    g <- function(y)
    {
        b <- 1; for(i in 1:3) 1 + a + b + aglobal + bglobal + x + y
    }
    for(i in 1:3) 1 + x + g(x)
}
test(f220, -1)
f230 <- function(x)    # same as f220 above but force promise
{
    jit(jit.flag, trace.flag)
    a <- 9
    b <- 9.1
    bglobal <- 9.2
    x <- x # force promise
    g <- function(y)
    {
        b <- 1;
        for(i in 1:3)
            2 + a + b + aglobal + bglobal + x + y
    }
    for(i in 1:3) 2 + x + g(x)
}
test(f230, -2)
f240 <- function(x)    # lexical scoping 2
{
    a <- 9
    b <- 9.1
    bglobal <- 9.2
    g <- function(y)
    {
        jit(jit.flag, trace.flag);
        b <- 1;
        for(i in 1:3)
            3 + a + b + aglobal + bglobal + x + y
    }
    for(i in 1:3) 3 + x + g(x)
}
test(f240, -3)
f250 <- function(x)    # same as f220 above but force promise
{
    a <- 9
    b <- 9.1
    bglobal <- 9.2
    x <- x # force promise
    g <- function(y)
    {
        jit(jit.flag, trace.flag);
        b <- 1;
        for(i in 1:3)
            4 + a + b + aglobal + bglobal + x + y
    }
    for(i in 1:3) 4 + x + g(x)
}
test(f250, -4)
f260 <- function(x)
{
    g <- function(y) { jit(1, 1); for(i in 1:3) x + y }
    g(3) + 3
}
test(f260, -4)  # RA_TODO why does x+y not get jitted here?
f262 <- function(x)
{
    g <- function(y) { jit(1, 1); for(i in 1:3) x + y + 1 }
    g(3) + 3
}
test(f262, -4)  # RA_TODO why only x+y jitted not x+y+1 here?
f270 <- function(x)     # test invocation of some common R functions from a jit block
{
    jit(jit.flag, trace.flag)
    library(mgcv)
    data(cars)
    count <- 0
    set.seed(1)
    x0 <- runif(30, 0, 1)
    x1 <- runif(30, 0, 1)
    for(i in 1:x) {
        count <- count + 1
        lmodel <- summary(lm(speed~dist, data=cars))
        if(i == 2) {
            cat("The answer is\n")
            print(lmodel)
        }
        y <- sin(pi * x0) + x1
        gmodel <- summary(gam(y~s(x0)+s(x1)))
   }
   c(lmodel$coefficients, lmodel$residuals, gmodel$coefficients, gmodel$residuals)
}
test(f270, 2)

#------------------------------------------------------------------------------
# closures
# RA_TODO the ugly "+ 0 + 0" code is a work around for the bug above in f260

return.incA <- function(inc)
{
    jit(0)
    function(x) { for(i in 1:3) x + inc + 0 + 0 }
}
incA.by.3 <- return.incA(3)
# following correctly won't get jitted because jit() is outside f
return.incA.jitted <- function(inc)
{
    jit(1, 1)
    function(x) { for(i in 1:3) x + inc + 0 + 0 }
}
incA.by.3.jitted <- return.incA.jitted(3)
if(incA.by.3(2.2) != incA.by.3.jitted(2.2))
    stop("incA.by.3.jitted failed");

return.incB <- function(inc)
{
    function(x) { jit(0); for(i in 1:3) x + inc + 0 + 0 }
}
incB.by.3 <- return.incB(3)
return.incB.jitted <- function(inc)
{
    function(x) { jit(1, 1); for(i in 1:3) x + inc + 0 + 0 }
}
incB.by.3.jitted <- return.incB.jitted(3)
if(incB.by.3(2.2) != incB.by.3.jitted(2.2))
    stop("incB.by.3.jitted failed");

#------------------------------------------------------------------------------
# subscripts

f300 <- function(...) # test subscripts, basic test
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3)
        x[i] <- i     # no jit compilation for this
    x
}
test(f300)
f310 <- function(...) # test subscripts, basic test
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3)
        x[i] <- xglobal[i] + 1 + 99
    x
}
test(f310)
f320 <- function(...) # test calculated subscript
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3)
        x[i+1] <- x[i] * 2
    x
}
test(f320)
f330 <- function(...) # test subscripts, vector expression
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:3)
        x <- x * x[i]
    x
}
test(f330)
f340 <- function(...) # test subscripts, vector expressions
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    y <- c(1,2)
    for(i in 1:3) {
         x[2:3] <- x[1] + 1
         x[3:4] <- x[2] + y
         x[4:5] <- x[3] + c(1,2)
         x[1:2] <- x[4] + i
    }
    x
}

#------------------------------------------------------------------------------
# binary ops

f400 <- function(...) # binops: scalar+scalar
{
    jit(jit.flag, trace.flag)
    ans <- double(28)
    for(i in 1:2) {
        ans[ 1] <- 2 + 3  # r1 r1
        ans[ 2] <- 2 - 3
        ans[ 3] <- 2 * 3
        ans[ 4] <- 2 / 3
        ans[ 5] <- 2 ^ 3
        ans[ 6] <- 2 %% 3
        ans[ 7] <- 2 %/% 3

        ans[ 8] <- i + 3  # i1 r1
        ans[ 9] <- i - 3
        ans[10] <- i * 3
        ans[11] <- i / 3
        ans[12] <- i ^ 3
        ans[13] <- i %% 3
        ans[14] <- i %/% 3

        ans[15] <- 2 + i # r1 i1
        ans[16] <- 2 - i
        ans[17] <- 2 * i
        ans[18] <- 2 / i
        ans[19] <- 2 ^ i
        ans[20] <- 2 %% i
        ans[21] <- 2 %/% i

        ans[22] <- i + i # i1 i1
        ans[23] <- i - i
        ans[24] <- i * i
        ans[25] <- i / i
        ans[26] <- i ^ i
        ans[27] <- i %% i
        ans[28] <- i %/% i
    }
    ans
}
test(f400)
f410 <- function(...) # binops: scalar+vector
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- 1:5

    for(i in 1:2) {
        ans[ 1,] <- 2 + x     # r1 r
        ans[ 2,] <- 2 - x
        ans[ 3,] <- 2 * x
        ans[ 4,] <- 2 / x
        ans[ 5,] <- 2 ^ x
        ans[ 6,] <- 2 %% x
        ans[ 7,] <- 2 %/% x

        ans[ 8,] <- i + x     # i1 r
        ans[ 9,] <- i - x
        ans[10,] <- i * x
        ans[11,] <- i / x
        ans[12,] <- i ^ x
        ans[13,] <- i %% x
        ans[14,] <- i %/% x

        ans[15,] <- 2 + xi   # r1 i
        ans[16,] <- 2 - xi
        ans[17,] <- 2 * xi
        ans[18,] <- 2 / xi
        ans[19,] <- 2 ^ xi
        ans[20,] <- 2 %% xi
        ans[21,] <- 2 %/% xi

        ans[22,] <- i + xi   # i1 i
        ans[23,] <- i - xi
        ans[24,] <- i * xi
        ans[25,] <- i / xi
        ans[26,] <- i ^ xi
        ans[27,] <- i %% xi
        ans[28,] <- i %/% xi
    }
    ans
}
test(f410)
f420 <- function(...) # binops: vector+scalar
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- 1:5

    for(i in 1:2) {
        ans[ 1,] <- x + 2     # r r1
        ans[ 2,] <- x - 2
        ans[ 3,] <- x * 2
        ans[ 4,] <- x / 2
        ans[ 5,] <- x ^ 2
        ans[ 6,] <- x %% 2
        ans[ 7,] <- x %/% 2

        ans[ 8,] <- xi + 2    # i r1
        ans[ 9,] <- xi - 2
        ans[10,] <- xi * 2
        ans[11,] <- xi / 2
        ans[12,] <- xi ^ 2
        ans[13,] <- xi %% 2
        ans[14,] <- xi %/% 2

        ans[15,] <- x + i    # r i1
        ans[16,] <- x - i
        ans[17,] <- x * i
        ans[18,] <- x / i
        ans[19,] <- x ^ i
        ans[20,] <- x %% i
        ans[21,] <- x %/% i

        ans[22,] <- xi + i   # i i1
        ans[23,] <- xi - i
        ans[24,] <- xi * i
        ans[25,] <- xi / i
        ans[26,] <- xi ^ i
        ans[27,] <- xi %% i
        ans[28,] <- xi %/% i
    }
    ans
}
test(f420)
f430 <- function(...) # binops: vector+vector
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- 1:5

    for(i in 1:2) {
        ans[ 1,] <- x + x     # r r
        ans[ 2,] <- x - x
        ans[ 3,] <- x * x
        ans[ 4,] <- x / x
        ans[ 5,] <- x ^ x
        ans[ 6,] <- x %% x
        ans[ 7,] <- x %/% x

        ans[ 8,] <- xi + x    # i r
        ans[ 9,] <- xi - x
        ans[10,] <- xi * x
        ans[11,] <- xi / x
        ans[12,] <- xi ^ x
        ans[13,] <- xi %% x
        ans[14,] <- xi %/% x

        ans[15,] <- x + xi   # r i
        ans[16,] <- x - xi
        ans[17,] <- x * xi
        ans[18,] <- x / xi
        ans[19,] <- x ^ xi
        ans[20,] <- x %% xi
        ans[21,] <- x %/% xi

        ans[22,] <- xi + xi  # i i
        ans[23,] <- xi - xi
        ans[24,] <- xi * xi
        ans[25,] <- xi / xi
        ans[26,] <- xi ^ xi
        ans[27,] <- xi %% xi
        ans[28,] <- xi %/% xi
    }
    ans
}
test(f430)

#------------------------------------------------------------------------------
# Binary ops with NA and div-by-zero
# We expect double NAs to be treated as NaNs whether you are jitting or not,
# but not integer NAs.

f402 <- function(...) # binops: scalar+scalar
{
    jit(jit.flag, trace.flag)
    ans <- double(46)
    ina <- as.integer(NA)
    rna <- as.double(NA)
    izero <- as.integer(0)
    rzero <- 0
    for(i in 1:2) {
        ans[ 1] <- 2 + rna  # r1 r1
        ans[ 2] <- 2 - rna
        ans[ 3] <- 2 * rna
        ans[ 4] <- 2 / rna
        ans[ 5] <- 2 ^ rna
        ans[ 6] <- 2 %% rna
        ans[ 7] <- 2 %/% rna

        ans[ 8] <- i + rna  # i1 r1
        ans[ 9] <- i - rna
        ans[10] <- i * rna
        ans[11] <- i / rna
        ans[12] <- i ^ rna
        ans[13] <- i %% rna
        ans[14] <- i %/% rna

        ans[15] <- rna + i # r1 i1
        ans[16] <- rna - i
        ans[17] <- rna * i
        ans[18] <- rna / i
        ans[19] <- rna ^ i
        ans[20] <- rna %% i
        ans[21] <- rna %/% i

        ans[22] <- i + ina # i1 i1
        ans[23] <- i - ina
        ans[24] <- i * ina
        ans[25] <- i / ina
        ans[26] <- i ^ ina
        ans[27] <- i %% ina
        ans[28] <- i %/% ina

        ans[29] <- 2 / rzero
        ans[30] <- i / rzero
        ans[31] <- rzero / i
        ans[32] <- i / izero
        ans[33] <- i %% izero
        ans[34] <- i %/% izero

        ans[35] <- 2 ^ rzero
        ans[36] <- i ^ rzero
        ans[37] <- rzero ^ i
        ans[38] <- i ^ izero
        ans[39] <- i %% izero
        ans[40] <- i %/% izero

        ans[41] <- rzero ^ 2
        ans[42] <- rzero ^ i
        ans[43] <- i ^ rzero
        ans[44] <- izero ^ i
        ans[45] <- izero %% i
        ans[46] <- izero %/% i
    }
    if(jit()[1]) {  # known issue
        ans[22] <- NA # was -2147483646
        ans[23] <- NA # was -2147483646
        ans[24] <- NA # was 0
        ans[25] <- NA # was -9.313226e-10
        ans[26] <- NA # was 0
    }
    ans
}
test(f402)
f412 <- function(...) # binops: scalar+vector
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=46,ncol=5)
    ina <- as.integer(NA)
    rna <- as.double(NA)
    izero <- as.integer(0)
    rzero <- 0
    x <- c(-9,-8,10,11,-12)
    xi <- 1:5
    x[1] <- rna
    x[2] <- rzero
    x[3] <- NaN
    xi[2] <- ina
    xi[3] <- izero
    xi[4] <- ina

    for(i in 1:2) {
        ans[ 1,] <- 2 + x     # r1 r
        ans[ 2,] <- 2 - x
        ans[ 3,] <- 2 * x
        ans[ 4,] <- 2 / x
        ans[ 5,] <- 2 ^ x
        ans[ 6,] <- 2 %% x
        ans[ 7,] <- 2 %/% x

        ans[ 8,] <- i + x     # i1 r
        ans[ 9,] <- i - x
        ans[10,] <- i * x
        ans[11,] <- i / x
        ans[12,] <- i ^ x
        ans[13,] <- i %% x
        ans[14,] <- i %/% x

        ans[15,] <- 2 + xi   # r1 i
        ans[16,] <- 2 - xi
        ans[17,] <- 2 * xi
        ans[18,] <- 2 / xi
        ans[19,] <- 2 ^ xi
        ans[20,] <- 2 %% xi
        ans[21,] <- 2 %/% xi

        ans[22,] <- i + xi   # i1 i
        ans[23,] <- i - xi
        ans[24,] <- i * xi
        ans[25,] <- i / xi
        ans[26,] <- i ^ xi
        ans[27,] <- i %% xi
        ans[28,] <- i %/% xi

        ans[29,] <- 2 / rzero
        ans[30,] <- i / rzero
        ans[31,] <- 2 / izero
        ans[32,] <- i / izero
        ans[33,] <- i %% izero
        ans[34,] <- i %/% izero

        ans[35,] <- 2 ^ rzero
        ans[36,] <- i ^ rzero
        ans[37,] <- 2 ^ izero
        ans[38,] <- i ^ izero
        ans[39,] <- i %% izero
        ans[40,] <- i %/% izero

        ans[41,] <- rzero ^ 2
        ans[42,] <- rzero ^ i
        ans[43,] <- izero ^ 2
        ans[44,] <- izero ^ i
        ans[45,] <- izero %% i
        ans[46,] <- izero %/% i
    }
    if(jit()[1]) {  # known issue
        ans[15,2] <- NA # was -2147483646
        ans[15,4] <- NA # was -2147483646
        ans[16,2] <- NA # was 2147483650
        ans[16,4] <- NA # was 2147483650
        ans[17,2] <- NA # was -4294967296
        ans[17,4] <- NA # was -4294967296
        ans[18,2] <- NA # was -9.313226e-10
        ans[18,4] <- NA # was -9.313226e-10
        ans[19,2] <- NA # was 0
        ans[19,4] <- NA # was 0
        ans[22,2] <- NA # was -2147483646
        ans[22,4] <- NA # was -2147483646
        ans[23,2] <- NA # was -2147483646
        ans[23,4] <- NA # was -2147483646
        ans[24,2] <- NA # was 0
        ans[24,4] <- NA # was 0
        ans[25,2] <- NA # was -9.313226e-10
        ans[25,4] <- NA # was -9.313226e-10
        ans[26,2] <- NA # was 0
        ans[26,4] <- NA # was 0
    }
    ans
}
test(f412)
f422 <- function(...) # binops: vector+scalar
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=46,ncol=5)
    ina <- as.integer(NA)
    rna <- as.double(NA)
    izero <- as.integer(0)
    rzero <- 0
    x <- c(-9,-8,10,11,-12)
    xi <- 1:5
    x[1] <- rna
    x[2] <- rzero
    x[3] <- NaN
    xi[2] <- ina
    xi[3] <- izero
    xi[4] <- ina

    for(i in 1:2) {
        ans[ 1,] <- x + 2     # r r1
        ans[ 2,] <- x - 2
        ans[ 3,] <- x * 2
        ans[ 4,] <- x / 2
        ans[ 5,] <- x ^ 2
        ans[ 6,] <- x %% 2
        ans[ 7,] <- x %/% 2

        ans[ 8,] <- xi + 2    # i r1
        ans[ 9,] <- xi - 2
        ans[10,] <- xi * 2
        ans[11,] <- xi / 2
        ans[12,] <- xi ^ 2
        ans[13,] <- xi %% 2
        ans[14,] <- xi %/% 2

        ans[15,] <- x + i    # r i1
        ans[16,] <- x - i
        ans[17,] <- x * i
        ans[18,] <- x / i
        ans[19,] <- x ^ i
        ans[20,] <- x %% i
        ans[21,] <- x %/% i

        ans[22,] <- xi + i   # i i1
        ans[23,] <- xi - i
        ans[24,] <- xi * i
        ans[25,] <- xi / i
        ans[26,] <- xi ^ i
        ans[27,] <- xi %% i
        ans[28,] <- xi %/% i

        ans[29,] <- x / rzero
        ans[30,] <- xi / rzero
        ans[31,] <- x / izero
        ans[32,] <- xi / izero
        ans[33,] <- xi %% izero
        ans[34,] <- xi %/% izero

        ans[35,] <- x ^ rzero
        ans[36,] <- xi ^ rzero
        ans[37,] <- x ^ izero
        ans[38,] <- xi ^ izero
        ans[39,] <- xi %% izero
        ans[40,] <- xi %/% izero

        ans[41,] <- rzero ^ x
        ans[42,] <- rzero ^ xi
        ans[43,] <- izero ^ x
        ans[44,] <- izero ^ xi
        ans[45,] <- izero %% xi
        ans[46,] <- izero %/% xi
    }
    # Known issue:
    #    standard R:    0 ^ rna <- Inf
    #    jitted:        0 ^ rna <- NA   tested with gcc 4.2.1-sjlj mingw32-2
    if(jit()[1]) {
        ans[41,1] <- Inf # was NA
        ans[41,3] <- Inf # was NaN
        ans[43,1] <- Inf # was NA
        ans[43,3] <- Inf # was NaN

        ans[8,2] <- NA # was -2147483646
        ans[8,4] <- NA # was -2147483646
        ans[9,2] <- NA # was -2147483650
        ans[9,4] <- NA # was -2147483650
        ans[10,2] <- NA # was -4294967296
        ans[10,4] <- NA # was -4294967296
        ans[11,2] <- NA # was -1073741824
        ans[11,4] <- NA # was -1073741824
        ans[12,2] <- NA # was 4.611686e+18
        ans[12,4] <- NA # was 4.611686e+18
        ans[22,2] <- NA # was -2147483646
        ans[22,4] <- NA # was -2147483646
        ans[23,2] <- NA # was 2147483646
        ans[23,4] <- NA # was 2147483646
        ans[24,2] <- NA # was 0
        ans[24,4] <- NA # was 0
        ans[25,2] <- NA # was -1073741824
        ans[25,4] <- NA # was -1073741824
        ans[26,2] <- NA # was 4.611686e+18
        ans[26,4] <- NA # was 4.611686e+18
        ans[30,2] <- NA # was -Inf
        ans[30,4] <- NA # was -Inf
        ans[32,2] <- NA # was -Inf
        ans[32,4] <- NA # was -Inf
        ans[38,2] <- NA # was 1
        ans[38,4] <- NA # was 1
        ans[44,2] <- NA # was Inf
        ans[44,4] <- NA # was Inf
    }
    ans
}
test(f422)
f432 <- function(...) # binops: vector+vector
{
    jit(jit.flag, trace.flag)
    ina <- as.integer(NA)
    rna <- as.double(NA)
    ans <- matrix(0,nrow=462,ncol=5)
    izero <- as.integer(0)
    rzero <- 0
    x <- c(-9,-8,10,11,-12)
    xi <- 1:5
    x[1] <- rna
    x[2] <- rzero
    x[3] <- NaN
    xi[2] <- ina
    xi[3] <- izero
    xi[4] <- ina

    for(i in 1:2) {
        ans[ 1,] <- x + x     # r r
        ans[ 2,] <- x - x
        ans[ 3,] <- x * x
        ans[ 4,] <- x / x
        ans[ 5,] <- x ^ x
        ans[ 6,] <- x %% x
        ans[ 7,] <- x %/% x

        ans[ 8,] <- xi + x    # i r
        ans[ 9,] <- xi - x
        ans[10,] <- xi * x
        ans[11,] <- xi / x
        ans[12,] <- xi ^ x
        ans[13,] <- xi %% x
        ans[14,] <- xi %/% x

        ans[15,] <- x + xi   # r i
        ans[16,] <- x - xi
        ans[17,] <- x * xi
        ans[18,] <- x / xi
        ans[19,] <- x ^ xi
        ans[20,] <- x %% xi
        ans[21,] <- x %/% xi

        ans[22,] <- xi + xi  # i i
        ans[23,] <- xi - xi
        ans[24,] <- xi * xi
        ans[25,] <- xi / xi
        ans[26,] <- xi ^ xi
        ans[27,] <- xi %% xi
        ans[28,] <- xi %/% xi

        ans[29,] <- x / rzero
        ans[30,] <- xi / rzero
        ans[31,] <- x / izero
        ans[32,] <- xi / izero
        ans[33,] <- xi %% izero
        ans[34,] <- xi %/% izero

        ans[35,] <- x ^ rzero
        ans[36,] <- xi ^ rzero
        ans[37,] <- x ^ izero
        ans[38,] <- xi ^ izero
        ans[39,] <- xi %% izero
        ans[40,] <- xi %/% izero

        ans[41,] <- rzero ^ x
        ans[42,] <- rzero ^ xi
        ans[43,] <- izero ^ x
        ans[44,] <- izero ^ xi
        ans[45,] <- izero %% xi
        ans[46,] <- izero %/% xi
    }
    # Known issue:
    #    standard R:    0 ^ rna <- Inf
    #    jitted:        0 ^ rna <- NA   tested with gcc 4.2.1-sjlj mingw32-2
    if(jit()[1]) {
        ans[12,3] <- Inf
        ans[41,1] <- Inf
        ans[41,3] <- Inf
        ans[43,1] <- Inf
        ans[43,3] <- Inf

        ans[8,2] <- NA # was -2147483648
        ans[8,4] <- NA # was -2147483644
        ans[9,2] <- NA # was -2147483648
        ans[9,4] <- NA # was -2147483652
        ans[10,2] <- NA # was 0
        ans[10,4] <- NA # was -8589934592
        ans[11,2] <- NA # was -Inf
        ans[11,4] <- NA # was -536870912
        ans[12,4] <- NA # was 2.126765e+37
        ans[15,2] <- NA # was -2147483648
        ans[15,4] <- NA # was -2147483644
        ans[16,2] <- NA # was 2147483648
        ans[16,4] <- NA # was 2147483652
        ans[17,2] <- NA # was 0
        ans[17,4] <- NA # was -8589934592
        ans[18,2] <- NA # was 0
        ans[18,4] <- NA # was -1.862645e-09
        ans[19,4] <- NA # was 0
        ans[22,2] <- NA # was 0
        ans[22,4] <- NA # was 0
        ans[23,2] <- NA # was 0
        ans[23,4] <- NA # was 0
        ans[24,2] <- NA # was 0
        ans[24,4] <- NA # was 0
        ans[25,2] <- NA # was 1
        ans[25,4] <- NA # was 1
        ans[26,2] <- NA # was 0
        ans[26,4] <- NA # was 0
        ans[30,2] <- NA # was -Inf
        ans[30,4] <- NA # was -Inf
        ans[32,2] <- NA # was -Inf
        ans[32,4] <- NA # was -Inf
        ans[38,2] <- NA # was 1
        ans[38,4] <- NA # was 1
        ans[44,2] <- NA # was Inf
        ans[44,4] <- NA # was Inf
    }

    ans
}
test(f432)

#------------------------------------------------------------------------------
# Binary ops with LOGICALs
# Remember, NA is logical

f404 <- function(...) # binops: scalar+scalar
{
    jit(jit.flag, trace.flag)
    ans <- double(28)
    for(i in 1:2) {
        ans[ 1] <- 2 + 3  # r1 r1
        ans[ 2] <- 2 - 3
        ans[ 3] <- 2 * 3
        ans[ 4] <- 2 / 3
        ans[ 5] <- 2 ^ 3
        ans[ 6] <- 2 %% 3
        ans[ 7] <- 2 %/% 3

        ans[ 8] <- TRUE + 3  # i1 r1
        ans[ 9] <- FALSE - 3
        ans[10] <- NA * 3
        ans[11] <- i / 3
        ans[12] <- i ^ 3
        ans[13] <- i %% 3
        ans[14] <- i %/% 3

        ans[15] <- 2 + TRUE # r1 i1
        ans[16] <- 2 - FALSE
        ans[17] <- 2 * NA
        ans[18] <- 2 / i
        ans[19] <- 2 ^ i
        ans[20] <- 2 %% i
        ans[21] <- 2 %/% i

        ans[22] <- TRUE + i # i1 i1
        ans[23] <- FALSE - i
        ans[24] <- NA * i
        ans[25] <- i / i
        ans[26] <- i ^ i
        ans[27] <- i %% i
        ans[28] <- i %/% i
    }
    if(jit()[1]) {      # known limitation: jit treats NA as INT_MIN
        ans[10] <- NA
        ans[17] <- NA
        ans[24] <- NA
    }

    ans
}
test(f404)
f414 <- function(...) # binops: scalar+vector
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- c(TRUE, FALSE, NA, TRUE, FALSE)

    for(i in 1:2) {
        ans[ 1,] <- 2 + x     # r1 r
        ans[ 2,] <- 2 - x
        ans[ 3,] <- 2 * x
        ans[ 4,] <- 2 / x
        ans[ 5,] <- 2 ^ x
        ans[ 6,] <- 2 %% x
        ans[ 7,] <- 2 %/% x

        ans[ 8,] <- i + x     # i1 r
        ans[ 9,] <- i - x
        ans[10,] <- i * x
        ans[11,] <- i / x
        ans[12,] <- i ^ x
        ans[13,] <- i %% x
        ans[14,] <- i %/% x

        ans[15,] <- 2 + xi   # r1 i
        ans[16,] <- 2 - xi
        ans[17,] <- 2 * xi
        ans[18,] <- 2 / xi
        ans[19,] <- 2 ^ xi
        ans[20,] <- 2 %% xi
        ans[21,] <- 2 %/% xi

        ans[22,] <- i + xi   # i1 i
        ans[23,] <- i - xi
        ans[24,] <- i * xi
        ans[25,] <- i / xi
        ans[26,] <- i ^ xi
        ans[27,] <- i %% xi
        ans[28,] <- i %/% xi
    }
    if(jit()[1]) {      # known limitation: logical NA treated as INT_MIN
        ans[15,3] <- NA
        ans[16,3] <- NA
        ans[17,3] <- NA
        ans[18,3] <- NA
        ans[19,3] <- NA
        ans[22,3] <- NA
        ans[23,3] <- NA
        ans[24,3] <- NA
        ans[25,3] <- NA
        ans[26,3] <- NA
    }
    ans
}
test(f414)
f424 <- function(...) # binops: vector+scalar
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- c(TRUE, FALSE, NA, TRUE, FALSE)

    for(i in 1:2) {
        ans[ 1,] <- x + 2     # r r1
        ans[ 2,] <- x - 2
        ans[ 3,] <- x * 2
        ans[ 4,] <- x / 2
        ans[ 5,] <- x ^ 2
        ans[ 6,] <- x %% 2
        ans[ 7,] <- x %/% 2

        ans[ 8,] <- xi + 2    # i r1
        ans[ 9,] <- xi - 2
        ans[10,] <- xi * 2
        ans[11,] <- xi / 2
        ans[12,] <- xi ^ 2
        ans[13,] <- xi %% 2
        ans[14,] <- xi %/% 2

        ans[15,] <- x + i    # r i1
        ans[16,] <- x - i
        ans[17,] <- x * i
        ans[18,] <- x / i
        ans[19,] <- x ^ i
        ans[20,] <- x %% i
        ans[21,] <- x %/% i

        ans[22,] <- xi + i   # i i1
        ans[23,] <- xi - i
        ans[24,] <- xi * i
        ans[25,] <- xi / i
        ans[26,] <- xi ^ i
        ans[27,] <- xi %% i
        ans[28,] <- xi %/% i
    }
    if(jit()[1]) {      # known limitation: jit treats NA as INT_MIN
        ans[8,3] <- NA
        ans[9,3] <- NA
        ans[10,3] <- NA
        ans[11,3] <- NA
        ans[12,3] <- NA
        ans[22,3] <- NA
        ans[23,3] <- NA
        ans[24,3] <- NA
        ans[25,3] <- NA
        ans[26,3] <- NA
    }
    ans
}
test(f424)
f434 <- function(...) # binops: vector+vector
{
    jit(jit.flag, trace.flag)
    ans <- matrix(0,nrow=28,ncol=5)
    x <- c(-9,0,10,11,-12)
    xi <- c(TRUE, FALSE, NA, TRUE, FALSE)

    for(i in 1:2) {
        ans[ 1,] <- x + x     # r r
        ans[ 2,] <- x - x
        ans[ 3,] <- x * x
        ans[ 4,] <- x / x
        ans[ 5,] <- x ^ x
        ans[ 6,] <- x %% x
        ans[ 7,] <- x %/% x

        ans[ 8,] <- xi + x    # i r
        ans[ 9,] <- xi - x
        ans[10,] <- xi * x
        ans[11,] <- xi / x
        ans[12,] <- xi ^ x
        ans[13,] <- xi %% x
        ans[14,] <- xi %/% x

        ans[15,] <- x + xi   # r i
        ans[16,] <- x - xi
        ans[17,] <- x * xi
        ans[18,] <- x / xi
        ans[19,] <- x ^ xi
        ans[20,] <- x %% xi
        ans[21,] <- x %/% xi

        ans[22,] <- xi + xi  # i i
        ans[23,] <- xi - xi
        ans[24,] <- xi * xi
        ans[25,] <- xi / xi
        ans[26,] <- xi ^ xi
        ans[27,] <- xi %% xi
        ans[28,] <- xi %/% xi
    }
    if(jit()[1]) {      # known limitation: jit treats NA as INT_MIN
        ans[8,3] <- NA
        ans[9,3] <- NA
        ans[10,3] <- NA
        ans[11,3] <- NA
        ans[12,3] <- NA
        ans[15,3] <- NA
        ans[16,3] <- NA
        ans[17,3] <- NA
        ans[18,3] <- NA
        ans[19,3] <- NA
        ans[22,3] <- NA
        ans[23,3] <- NA
        ans[24,3] <- NA
        ans[25,3] <- NA
        ans[26,3] <- NA
    }
    ans
}
test(f434)

#------------------------------------------------------------------------------
# RA_TODO test integer NAs in math1 functions

f500 <- function(...) # test math1_r1
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:5) {
        x[i] <- sin(xglobal[i])
        x[i] <- x[i] + xglobal[i] + 1 + cos(xglobal[i])
        x[i] <- x[i] + tan(xglobal[i]) + xglobal[i] + 1
    }
    x
}
test(f500)
f510 <- function(...) # test math1_i1
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:5) {
        x[i] <- sin(i)
        x[i] <- x[i] + xglobal[i] + 1 + cos(i)
        x[i] <- x[i] + tan(i) + xglobal[i] + 1
    }
    x
}
test(f510)
f530 <- function(...) # test math1_r
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    for(i in 1:5) {
        x <- xglobal[i] + 1 + cos(xglobal)
        x <- x + tan(xglobal) + xglobal[i] + 1
    }
    x
}
test(f530)
f540 <- function(...) # test math1_i
{
    jit(jit.flag, trace.flag)
    x <- c(-9,0,10,11,-12)
    xi <- 1:5
    for(i in 1:5) {
        x <- xglobal[i] + 1 + cos(xi)
        x <- x + tan(xi) + xglobal[i] + 1
    }
    x
}
test(f540)

#------------------------------------------------------------------------------
# error handling: these deliberately cause an error which we catch in a try block

test.try <- function(f) # for testing that jitting works in f when invoked by try
{
    z <- try(f(), silent=TRUE)
    if(class(z) == "try-error")
        stop("test ", paste(substitute(f)), " failed")
}

f600 <- function() # should cause no error, test with test.try
{
    jit(1,1)
    x <- 0
    for(i in 1:3)
        x <- x+1

}
test.try(f600)  # this is a test of running jitted code under try()

test.err.handling <- function(f)    # for testing that error caused by f is caught
{
    z <- try(f(), silent=TRUE)
    cat("Expect an err message right here-->", gsub("\n", " ", z[1]), "\n");
    if(class(z) != "try-error")
        stop("test ", paste(substitute(f)), " failed")
}
f601 <- function(...) # cannot change the type of a jitted variable
{
    jit(1, 1)
    x <- 0
    for(i in 1:3)
        if (i == 2)
            x <- i # double to int
        else
            x <- 2
    x
}
test.err.handling(f601)
f602 <- function(...)
{
    jit(jit.flag, trace.flag)
    x <- 0
    for(i in 1:3) {
        # The following statement LOOKS LIKE it is changing
        # the type but R changes the lhs to int.
        # Hence we use test() not test.err.handling() below.

        x <- i       # double to int
    }
    x
}
test(f602)
f603 <- function(...) # cannot change the type of a jitted variable
{
    jit(1, 1)
    x <- 0
    for(i in 1:3) {
        # this is not compiled (because atan2 is math2) but we _try_
        # to compile it and therefore the x gets marked as jitted.
        x <- atan2(2,3) + 1;
    }
    x <- "c"
    x
}
test.err.handling(f603)
f604 <- function() # ok to change type of jitted variable after jit(0)
{
    foo <- function() {
        jit(1, 1)
        y <- 0
        for(i in 1:3)
            y <- x + 1  # x gets jitted here
    }   # return does implicit jit(0) here
    x <- 1
    foo()
    x <- "c"
}
test(f604)
f605 <- function() # ok to change type of jitted variable after jit(0)
{
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 1:3)
        x <- x + 1    # x gets jitted here
    jit(0)
    x <- "c"
}
test(f605)
# this caused a bogus error message in print(system.time(foo()))
# before I added the clear_jitted_bits code
f605a <- function()
{
    foo <- function() 3
    jit(jit.flag, trace.flag)
    x <- 1
    for(i in 1:3)
       for(j in 1:3)
          x <- 2
    jit(0)
    print(system.time(foo()))
    x
}
test(f605a)
f606 <- function(...) # cannot change the type of a jitted variable, nested, arg
{
    foo <- function(x) {
        jit(1, 1)
        y <- 0
        for(i in 1:3) {
            y <- x + 11 + 12 + 13
            x <- 21:23
        }
    }
    foo(44)
}
test.err.handling(f606)
f607 <- function() # cannot change the length of a jitted variable
{
    jit(1,1)
    x <- 2
    for(i in 1:3)
        x <- x + 2
    length(x) <- 9
}
test.err.handling(f607)
f607a <- function() # cannot change the length of a jitted variable
{
    jit(1,1)
    x <- 1
    for(i in 1:3)
        x <- x + 1
    x <- c(1,2,3)  # [1] to [3]
}
test.err.handling(f607a)
f608_global <- 3
f608 <- function() # cannot change jitted variable from non-local to local scope
{
    jit(1,1)
    x <- 0
    for(i in 1:3) {
        x <- 1 + f608_global
        f608_global <- 4
    }
    x
}
test.err.handling(f608)
f610 <- function() # cannot remove a jitted variable
{
    jit(1,1)
    for(i in 1:3)
        x <- x + 1
    remove(x)
}
test.err.handling(f610)
f612 <- function() # too many expressions in jit block, compare to f050
{
    if (trace.flag == 1)
        trace.flag = 2; 	# use trace=2 so we see "did not compile" msgs
    jit(jit.flag, trace.flag)
    cat("jit()=", jit(), "\n")
    x <- 0
    for(i in 1:3) {
        x <- x +  1; x <- x +  2; x <- x +  3; x <- x +  4; x <- x +  5;
        x <- x +  6; x <- x +  7; x <- x +  8; x <- x +  9; x <- x + 10;
        x <- x + 11; x <- x + 12; x <- x + 13; x <- x + 14; x <- x + 15;
        x <- x + 16; x <- x + 17; x <- x + 18; x <- x + 19; x <- x + 20;
        x <- x + 21; x <- x + 22; x <- x + 23; x <- x + 24; x <- x + 25;
        x <- x + 26; x <- x + 27; x <- x + 28; x <- x + 29; x <- x + 30;
        x <- x + 31; x <- x + 32; x <- x + 33; x <- x + 34; x <- x + 35;
        x <- x + 36; x <- x + 37; x <- x + 38; x <- x + 39; x <- x + 40;
        x <- x + 41; x <- x + 42; x <- x + 43; x <- x + 44; x <- x + 45;
        x <- x + 46; x <- x + 47; x <- x + 48; x <- x + 49; x <- x + 50;
        x <- x + 51; x <- x + 52; x <- x + 53; x <- x + 54; x <- x + 55;
        x <- x + 56; x <- x + 57; x <- x + 58; x <- x + 59; x <- x + 60;
        x <- x + 61; x <- x + 62; x <- x + 63; x <- x + 64; x <- x + 65;
        x <- x + 66; x <- x + 67; x <- x + 68; x <- x + 69; x <- x + 70;
        x <- x + 71; x <- x + 72; x <- x + 73; x <- x + 74; x <- x + 75;
        x <- x + 76; x <- x + 77; x <- x + 78; x <- x + 79; x <- x + 80;
        x <- x + 81; x <- x + 82; x <- x + 83; x <- x + 84; x <- x + 85;
        x <- x + 86; x <- x + 87; x <- x + 88; x <- x + 89; x <- x + 90;
        x <- x + 91; x <- x + 92; x <- x + 93; x <- x + 94; x <- x + 95;
        x <- x + 96; x <- x + 97; x <- x + 98; x <- x + 99; x <- x + 100;
        x <- x + 101    # the last straw
    }
    x
}
test(f612)
f618 <- function() # out of range index %d (allowed range is 1 to %d)
{
    jit(1,1)
    x <- 1:3
    y <- 0
    for(i in 1:3)
        y <- y - x[i+1]
    y
}
test.err.handling(f618)
f620 <- function() # out of range index %g (allowed range is 1 to %d)
{
    jit(1,1)
    x <- 1:3
    y <- 0
    for(i in 3:1)
        y <- y - x[i-1]
    y
}
test.err.handling(f620)
f628 <- function() # disallow if jitting: make an active binding
{
    jit(1,1)
    f <- function(x) x+1
    jit(1,1)
    makeActiveBinding("fred", f, .GlobalEnv)
}
test.err.handling(f628)

# RA_TODO not tested:
#     cannot unbind jitted variable '%s'
#     cannot use a '...' variable in a jit block
#     disallow if jitting: use active values
#     disallow if jitting: use user defined databases
#     disallow if jitting: attach
#     disallow if jitting: detach
#     disallow if jitting: use mkUnbound
#     disallow if jitting: libfixup
#     locked environments

#------------------------------------------------------------------------------
# Test on "real code".  This was copied from maptools 0.6-19 and hacked on.
#
# RA_TODO This will probably be removed but for now it is a convenient
# way of seeing what changes you might have to make to real code to get
# it to run faster with jit.  Not too much gets compiled here because
# most variables are complex.

pointLabel <- function(x, y = NULL, labels = seq(along = x), cex = 1,
                       method = c("SANN", "GA"),
                       allowSmallOverlap = FALSE,
                       trace = FALSE,
                       doPlot = FALSE,
                       ...)
{
    toUnityCoords <- function(xy) {
	    list(x = (xy$x - boundary[1]) / (boundary[2] - boundary[1]) * xyAspect,
	    y = (xy$y - boundary[3]) / (boundary[4] - boundary[3]) / xyAspect)
    }
    toUserCoords <- function(xy) {
	list(x = boundary[1] + xy$x / xyAspect * (boundary[2] - boundary[1]),
	y = boundary[3] + xy$y * xyAspect * (boundary[4] - boundary[3]))
    }
    if (!missing(y) && (is.character(y) || is.expression(y))) {
	labels <- y
	y <- NULL
    }
    gen_offset <- function(code) {
        c(-1,  -1,  -1,  0,  0,   1,  1,   1)[code] * (width/2) +
        1i * c(-1,   0,   1, -1,  1,  -1,  0,   1)[code] * (height/2)
    }
    rect_intersect <- function(xy1, offset1, xy2, offset2) {
	    w <- pmin(Re(xy1+offset1/2), Re(xy2+offset2/2)) - pmax(Re(xy1-offset1/2), Re(xy2-offset2/2))
	    h <- pmin(Im(xy1+offset1/2), Im(xy2+offset2/2)) - pmax(Im(xy1-offset1/2), Im(xy2-offset2/2))
	    w[w <= 0] <- 0
	    h[h <= 0] <- 0
	    w*h
    }
    nudge <- function(offset, rectidx1, rectidx2) {
	doesIntersect <- rect_intersect(xy[rectidx1] + offset[rectidx1], rectv[rectidx1],
			xy[rectidx2] + offset[rectidx2], rectv[rectidx2]) > 0

	pyth <- abs(xy[rectidx1] + offset[rectidx1] - xy[rectidx2] - offset[rectidx2]) / nudgeFactor
	eps <- 1.0e-10

	for (i in which(doesIntersect & pyth > eps)) {
		idx1 <- rectidx1[i]
		idx2 <- rectidx2[i]
		vect <- (xy[idx1] + offset[idx1] - xy[idx2] - offset[idx2]) / pyth[idx1]
		offset[idx1] <- offset[idx1] + vect
		offset[idx2] <- offset[idx2] - vect
	}
	offset
    }
    SANN <- function(rectidx1, rectidx2) {
	i <- 1; j <- 1 # prevent jit "cannot change jitted symbol 'i' to local scope"
	gene <- rep(8, n_labels)

	# following code was offset <- objective(newgene, rectidx1, rectidx2)
	offset <- gen_offset(gene)
	if (allowSmallOverlap) 
            offset <- nudge(offset, rectidx1, rectidx2)
	if (!is.null(rectidx1))
	    area <- sum(rect_intersect(xy[rectidx1] + offset[rectidx1], rectv[rectidx1],
		        xy[rectidx2] + offset[rectidx2], rectv[rectidx2]))
	else
	    area <- 0
 	n_outside <- sum(Re(xy + offset - rectv/2) < 0 | Re(xy + offset + rectv/2) > xyAspect |
 	                 Im(xy + offset - rectv/2) < 0 | Im(xy + offset + rectv/2) > 1/xyAspect)

	score <- 1000 * area + n_outside
	# end was

	bestgene <- gene
	bestscore <- score
	T <- 2.5
	N1 = 50
        jit(jit.flag, trace.flag)
	for (i in 1:N1) {
	    k <- 1
	    for (j in 1:N1) {
		newgene <- gene
                # RA_TODO get sample return mismatches unless we use a set.seed below -- why?
                set.seed(1000 * i + j)
 		newgene[sample(1:n_labels, 1)] <- sample(1:8,1)
 		# following code was newscore <- objective(newgene, rectidx1, rectidx2)
 		offset <- gen_offset(gene)
 		if (allowSmallOverlap) 
                     offset <- nudge(offset, rectidx1, rectidx2)
 		if (!is.null(rectidx1))
 		    area <- sum(rect_intersect(xy[rectidx1] + offset[rectidx1], rectv[rectidx1],
 				               xy[rectidx2] + offset[rectidx2], rectv[rectidx2]))
 		else
 		    area <- 0

		n_outside <- sum(Re(xy + offset - rectv/2) < 0 | Re(xy + offset + rectv/2) > xyAspect |
		                 Im(xy + offset - rectv/2) < 0 | Im(xy + offset + rectv/2) > 1/xyAspect)

 		newscore <- 1000 * area + n_outside
 		# end was
 
 		if (newscore <= score || runif(1) < exp((score - newscore) / T)) {
 		    k <- k + 1
 		    score <- newscore
 		    gene <- newgene
 		}
 		if (score <= bestscore) {
 		    bestscore <- score
 		    bestgene <- gene
 		}
 		if (bestscore == 0 || k == 10)
                # added "if" so we iterate at least a few times before break
                # this messes up the algorithm, it doesn't matter
                    if (i > 8 || N1 < 8)
                         break
	    }
	    if (bestscore == 0) 
                break
	    if (trace) cat("overlap area =", bestscore, "\n")
	    T <- 0.9 * T
	}
        jit(0)
	if (trace) cat("overlap area =", bestscore, "\n")
	nx <- Re(xy + gen_offset(bestgene))
	ny <- Im(xy + gen_offset(bestgene))
	list(x = nx, y = ny)
    }
    labels <- as.graphicsAnnot(labels)
    # changed following two lines to prevent graphics window display
    boundary <- c(0,1,0,1) # was par()$usr
    xyAspect <- 1.1 # was par()$pin[1] / par()$pin[2] # width / height
    z <- xy.coords(x, y, recycle = TRUE)
    z <- toUnityCoords(z)
    x <- z$x
    y <- z$y
    if (length(labels) < length(x))
        labels <- rep(labels, length(x))
    method <- match.arg(method)
    if (allowSmallOverlap)
        nudgeFactor <- 0.02
    n_labels <- length(x)
    # changed following two lines to prevent graphics window display
    width <- rep(0.0373, n_labels)  # was (strwidth(labels, units = "figure", cex = cex) + 0.015) * xyAspect
    height <- rep(0.0242, n_labels) # was (strheight(labels, units = "figure", cex = cex) + 0.015) / xyAspect
    xy <- x + 1i * y
    rectv <- width + 1i * height
    rectidx1 <- rectidx2 <- array(0, (length(x)^2 - length(x)) / 2)
    k <- 0
    jit(jit.flag, trace.flag)
    for (i in 1:length(x))
	for (j in seq(len=(i-1))) {
	    k <- k + 1
	    rectidx1[k] <- i
	    rectidx2[k] <- j
	}
    jit(0)
    canIntersect <- rect_intersect(xy[rectidx1], 2 * rectv[rectidx1],
		                   xy[rectidx2], 2 * rectv[rectidx2]) > 0
    rectidx_1 <- rectidx1[canIntersect] # prevent jit "cannot change len"
    rectidx_2 <- rectidx2[canIntersect]
    if (trace) cat("possible intersects =", length(rectidx1), "\n")
    if (trace) cat("portion covered =", sum(rect_intersect(xy, rectv,xy,rectv)),"\n")
    if (method == "SANN")
        xy <- SANN(rectidx_1, rectidx_2)
    else
        stop("only SANN is supported")
    xy <- toUserCoords(xy)
    if (doPlot)
        stop("doPlot not supported")
    invisible(xy)
}
N <- 5 # needs to be bigger for a valid comparison but that takes too long for routine testing
x <- runif(N)*10
y <- runif(N)*10
jit.flag = 0; trace.flag <- 0
set.seed(1)   # for reproducibility
time1 <- system.time(labels1 <- 
            pointLabel(x, y, method=c("SANN"), as.character(round(x,1)), offset=0, cex=.6))
# must do it again for valid time comparison (first time times are different --- why?)
set.seed(1)   # for reproducibility
time1 <- system.time(labels1 <- 
            pointLabel(x, y, method=c("SANN"), as.character(round(x,1)), offset=0, cex=.6))
jit.flag = 1; trace.flag <- 1
set.seed(1)
time2 <- system.time(labels2 <- pointLabel(x, y, method=c("SANN"), as.character(round(x,1)), offset=0, cex=.6))
# cat("SANN relative time ", format(100 * time2[1] / time1[1], digits=2), "%\n", sep="")

for (i in 1:length(labels1$x))
    if (!identical(labels1$x[i], labels2$x[i]) || !identical(labels1$y[i],labels2$y[i])) {
         cat("mismatch at [", i, "] labels1$x ", labels1$x[i], 
		" labels2$x ", labels2$x[i], 
		" labels1$y ", labels1$y[i], 
		" labels2$y ", labels2$y[i], "\n", sep="")
         stop("mismatch detected in pointLabel")
    }
stopifnot(identical(labels1, labels2))

#------------------------------------------------------------------------------

cat("Done\n")
