# time-jit.R

# dummy function for testing with standard R
#
# jit <- function(jit=NA, show=0, check=1) 
# {
#     if (length(grep("ra", R.version.string)) > 0)
# 	stop("Cannot redefine \"jit()\" when running ra")
#     c(jit, show, check, jit)
# }

jit.flag <- 0
show.flag <- 0

test <- function(f, N)
{
    strip.white.space <- function(s) gsub("[ \t\n]", "", s)

    jit.flag <<- 0; show.flag <<- 0; atime = system.time(a <- f(N));
    jit.flag <<- 1; show.flag <<- 0; btime = system.time(b <- f(N));

    if (!identical(a,b))
        stop("jit(1,1) != jit(0)")

    cat(sprintf("%-20.20s %4.0f%% of %4.2f secs, N=%g\n",
        strip.white.space(paste(substitute(f))),
        100 * btime[1]/atime[1],
        atime[1], N))
}
# # Tests from Vadim Ogranovich post.
# # These are pretty useless in terms of representing what people
# # want to actually do with R code.
# # Not much here gets jitted yet
# 
# vadim1.x_index_i_eq_x_index_i_minus_1 <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         x[i] = x[i-1]
#     x
# }
# vadim2.x_index_i_minus_1 <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         x[i-1]
#     x
# }
# vadim3.1 <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         1
#     x
# }
# vadim4.x_index_i_eq_1.0 <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         x[i] = 1.0
#     x
# }
# vadim5.i_minus_1 <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         i-1
#     x
# }
# vadim6.i <- function(N)
# {
#     jit(jit.flag, show.flag)
#     iA = seq(2,N); x = double(N); x[1] = 99; x[2] = 98
#     for (i in iA)
#         i
#     x
# }
# test(vadim1.x_index_i_eq_x_index_i_minus_1, N/20)
# test(vadim2.x_index_i_minus_1, N/4)
# test(vadim3.1, N*2)
# test(vadim4.x_index_i_eq_1.0, N/8)
# test(vadim5.i_minus_1, N/2)
# test(vadim6.i, N*2)

convolve <- function(N)
{
    jit(jit.flag, show.flag)
    a <- double(N)
    b <- double(N)
    na <- length(a)
    nb <- length(b)
    ab <- double(na + nb)
    for(i in 1:na)
        for(j in 1:nb)
             ab[i + j] <- ab[i + j] + a[i] * b[j]
    ab
}
simple.count <- function(N)
{
    jit(jit.flag, show.flag)
    x <- 0
    for(i in 1:N)
	x <- x+1
}
simple.subscript.assign <- function(N)
{
    jit(jit.flag, show.flag)
    x <- double(N)
    for(i in 1:N)
	x[i] <- 0
}
test(convolve, 400)  # 400 for quick, 600 for reliable
N = as.integer(2e6)  # 2e6 for quick, 1e7 for reliable
# we adjust N below so each test takes about same time
test(simple.count, as.integer(N/3))
test(simple.subscript.assign, as.integer(N/10))
